@@ -2,21 +2,19 @@
 import functools
 import random
 import time
+import warnings
 from collections import namedtuple
 
+from py_zipkin import Encoding
+from py_zipkin import Kind
+from py_zipkin import storage
+from py_zipkin.encoding._helpers import create_endpoint
+from py_zipkin.encoding._helpers import SpanBuilder
 from py_zipkin.exception import ZipkinError
-from py_zipkin.logging_helper import zipkin_logger
-from py_zipkin.logging_helper import ZipkinLoggerHandler
 from py_zipkin.logging_helper import ZipkinLoggingContext
-from py_zipkin.thread_local import get_zipkin_attrs
-from py_zipkin.thread_local import pop_zipkin_attrs
-from py_zipkin.thread_local import push_zipkin_attrs
-from py_zipkin.thrift import SERVER_ADDR_VAL
-from py_zipkin.thrift import create_binary_annotation
-from py_zipkin.thrift import create_endpoint
-from py_zipkin.thrift import zipkin_core
-from py_zipkin.util import generate_random_64bit_string
+from py_zipkin.storage import ThreadLocalStack
 from py_zipkin.util import generate_random_128bit_string
+from py_zipkin.util import generate_random_64bit_string
 
 
 """
@@ -33,12 +31,7 @@ ZipkinAttrs = namedtuple(
     ['trace_id', 'span_id', 'parent_span_id', 'flags', 'is_sampled'],
 )
 
-
-STANDARD_ANNOTATIONS = {
-    'client': {'cs', 'cr'},
-    'server': {'ss', 'sr'},
-}
-STANDARD_ANNOTATIONS_KEYS = frozenset(STANDARD_ANNOTATIONS.keys())
+ERROR_KEY = 'error'
 
 
 class zipkin_span(object):
@@ -109,11 +102,18 @@ class zipkin_span(object):
         binary_annotations=None,
         port=0,
         sample_rate=None,
-        include=('client', 'server'),
+        include=None,
         add_logging_annotation=False,
         report_root_timestamp=False,
         use_128bit_trace_id=False,
-        host=None
+        host=None,
+        context_stack=None,
+        span_storage=None,
+        firehose_handler=None,
+        kind=None,
+        timestamp=None,
+        duration=None,
+        encoding=Encoding.V1_THRIFT,
     ):
         """Logs a zipkin span. If this is the root span, then a zipkin
         trace is started as well.
@@ -126,9 +126,9 @@ class zipkin_span(object):
         :type zipkin_attrs: ZipkinAttrs
         :param transport_handler: Callback function that takes a message parameter
             and handles logging it
-        :type transport_handler: function
+        :type transport_handler: BaseTransportHandler
         :param max_span_batch_size: Spans in a trace are sent in batches,
-        max_span_batch_size defines max size of one batch
+            max_span_batch_size defines max size of one batch
         :type max_span_batch_size: int
         :param annotations: Optional dict of str -> timestamp annotations
         :type annotations: dict of str -> int
@@ -145,7 +145,8 @@ class zipkin_span(object):
         :type sample_rate: float
         :param include: which annotations to include
             can be one of {'client', 'server'}
-            corresponding to ('cs', 'cr') and ('ss', 'sr') respectively
+            corresponding to ('cs', 'cr') and ('ss', 'sr') respectively.
+            DEPRECATED: use kind instead. `include` will be removed in 1.0.
         :type include: iterable
         :param add_logging_annotation: Whether to add a 'logging_end'
             annotation when py_zipkin finishes logging spans
@@ -158,11 +159,31 @@ class zipkin_span(object):
             sampling decisions (i.e. are the root spans of entire traces) will
             always report timestamp/duration.
         :type report_root_timestamp: boolean
-        :param use_128bit_trace_id: If true, generate 128-bit trace_ids
+        :param use_128bit_trace_id: If true, generate 128-bit trace_ids.
         :type use_128bit_trace_id: boolean
-        :param host: Contains the ipv4 value of the host. The ipv4 value isn't
-            automatically determined in a docker environment
+        :param host: Contains the ipv4 or ipv6 value of the host. The ip value
+            isn't automatically determined in a docker environment.
         :type host: string
+        :param context_stack: explicit context stack for storing
+            zipkin attributes
+        :type context_stack: object
+        :param span_storage: explicit Span storage for storing zipkin spans
+            before they're emitted.
+        :type span_storage: py_zipkin.storage.SpanStorage
+        :param firehose_handler: [EXPERIMENTAL] Similar to transport_handler,
+            except that it will receive 100% of the spans regardless of trace
+            sampling rate.
+        :type firehose_handler: BaseTransportHandler
+        :param kind: Span type (client, server, local, etc...).
+        :type kind: Kind
+        :param timestamp: Timestamp in seconds, defaults to `time.time()`.
+            Set this if you want to use a custom timestamp.
+        :type timestamp: float
+        :param duration: Duration in seconds, defaults to the time spent in the
+            context. Set this if you want to use a custom duration.
+        :type duration: float
+        :param encoding: Output encoding format, defaults to V1 thrift spans.
+        :type encoding: Encoding
         """
         self.service_name = service_name
         self.span_name = span_name
@@ -172,39 +193,72 @@ class zipkin_span(object):
         self.annotations = annotations or {}
         self.binary_annotations = binary_annotations or {}
         self.port = port
-        self.logging_context = None
         self.sample_rate = sample_rate
-        self.include = include
         self.add_logging_annotation = add_logging_annotation
         self.report_root_timestamp_override = report_root_timestamp
         self.use_128bit_trace_id = use_128bit_trace_id
         self.host = host
-        self.logging_configured = False
+        self._context_stack = context_stack or ThreadLocalStack()
+        if span_storage is not None:
+            self._span_storage = span_storage
+        else:
+            self._span_storage = storage.default_span_storage()
+        self.firehose_handler = firehose_handler
+        self.kind = self._generate_kind(kind, include)
+        self.timestamp = timestamp
+        self.duration = duration
+        self.encoding = encoding
 
-        # Spans that log a 'cs' timestamp can additionally record
-        # 'sa' binary annotations that show where the request is going.
-        # This holds a list of 'sa' binary annotations.
-        self.sa_binary_annotations = []
+        self._is_local_root_span = False
+        self.logging_context = None
+        self.do_pop_attrs = False
+        # Spans that log a 'cs' timestamp can additionally record a
+        # 'sa' binary annotation that shows where the request is going.
+        self.sa_endpoint = None
 
-        # Validation checks
+        # It used to  be possible to override timestamp and duration by passing
+        # in the cs/cr or sr/ss annotations. We want to keep backward compatibility
+        # for now, so this logic overrides self.timestamp and self.duration in the
+        # same way.
+        # This doesn't fit well with v2 spans since those annotations are gone, so
+        # we also log a deprecation warning.
+        if 'sr' in self.annotations and 'ss' in self.annotations:
+            self.duration = self.annotations['ss'] - self.annotations['sr']
+            self.timestamp = self.annotations['sr']
+            warnings.warn(
+                "Manually setting 'sr'/'ss' annotations is deprecated. Please "
+                "use the timestamp and duration parameters.",
+                DeprecationWarning,
+            )
+        if 'cr' in self.annotations and 'cs' in self.annotations:
+            self.duration = self.annotations['cr'] - self.annotations['cs']
+            self.timestamp = self.annotations['cs']
+            warnings.warn(
+                "Manually setting 'cr'/'cs' annotations is deprecated. Please "
+                "use the timestamp and duration parameters.",
+                DeprecationWarning,
+            )
+
+        # Root spans have transport_handler and at least one of zipkin_attrs
+        # or sample_rate.
         if self.zipkin_attrs or self.sample_rate is not None:
+            # transport_handler is mandatory for root spans
             if self.transport_handler is None:
                 raise ZipkinError(
                     'Root spans require a transport handler to be given')
 
+            self._is_local_root_span = True
+
+        # If firehose_handler than this is a local root span.
+        if self.firehose_handler:
+            self._is_local_root_span = True
+
         if self.sample_rate is not None and not (0.0 <= self.sample_rate <= 100.0):
             raise ZipkinError('Sample rate must be between 0.0 and 100.0')
 
-        if not set(include).issubset(STANDARD_ANNOTATIONS_KEYS):
-            raise ZipkinError(
-                'Only %s are supported as annotations' %
-                STANDARD_ANNOTATIONS_KEYS
-            )
-        else:
-            # get a list of all of the mapped annotations
-            self.annotation_filter = set()
-            for include_name in include:
-                self.annotation_filter.update(STANDARD_ANNOTATIONS[include_name])
+        if not isinstance(self._span_storage, storage.SpanStorage):
+            raise ZipkinError('span_storage should be an instance '
+                              'of py_zipkin.storage.SpanStorage')
 
     def __call__(self, f):
         @functools.wraps(f)
@@ -218,8 +272,14 @@ class zipkin_span(object):
                 binary_annotations=self.binary_annotations,
                 port=self.port,
                 sample_rate=self.sample_rate,
-                include=self.include,
-                host=self.host
+                host=self.host,
+                context_stack=self._context_stack,
+                span_storage=self._span_storage,
+                firehose_handler=self.firehose_handler,
+                kind=self.kind,
+                timestamp=self.timestamp,
+                duration=self.duration,
+                encoding=self.encoding,
             ):
                 return f(*args, **kwargs)
         return decorated
@@ -227,6 +287,32 @@ class zipkin_span(object):
     def __enter__(self):
         return self.start()
 
+    def _generate_kind(self, kind, include):
+        # If `kind` is not set, then we generate it from `include`.
+        # This code maintains backward compatibility with old versions of py_zipkin
+        # which used include rather than kind to identify client / server spans.
+        if kind:
+            return kind
+        else:
+            if include:
+                # If `include` contains only one of `client` or `server`
+                # than it's a client or server span respectively.
+                # If neither or both are present, then it's a local span
+                # which is represented by kind = None.
+                warnings.warn(
+                    'The include argument is deprecated. Please use kind.',
+                    DeprecationWarning,
+                )
+                if 'client' in include and 'server' not in include:
+                    return Kind.CLIENT
+                elif 'client' not in include and 'server' in include:
+                    return Kind.SERVER
+                else:
+                    return Kind.LOCAL
+
+        # If both kind and include are unset, then it's a local span.
+        return Kind.LOCAL
+
     def start(self):
         """Enter the new span context. All annotations logged inside this
         context will be attributed to this span. All new spans generated
@@ -238,37 +324,71 @@ class zipkin_span(object):
         never attached in the unsampled case, so the spans are never logged.
         """
         self.do_pop_attrs = False
-        # If zipkin_attrs are passed in or this span is doing its own sampling,
-        # it will need to actually log spans at __exit__.
-        self.perform_logging = self.zipkin_attrs or self.sample_rate is not None
         report_root_timestamp = False
 
-        if self.sample_rate is not None:
-            if self.zipkin_attrs and not self.zipkin_attrs.is_sampled:
-                report_root_timestamp = True
-                self.zipkin_attrs = create_attrs_for_span(
-                    sample_rate=self.sample_rate,
-                    trace_id=self.zipkin_attrs.trace_id,
-                    use_128bit_trace_id=self.use_128bit_trace_id,
-                )
-            elif not self.zipkin_attrs:
-                report_root_timestamp = True
-                self.zipkin_attrs = create_attrs_for_span(
-                    sample_rate=self.sample_rate,
-                    use_128bit_trace_id=self.use_128bit_trace_id,
-                )
+        # This check is technically not necessary since only root spans will have
+        # sample_rate, zipkin_attrs or a transport set. But it helps making the
+        # code clearer by separating the logic for a root span from the one for a
+        # child span.
+        if self._is_local_root_span:
 
-        if not self.zipkin_attrs:
-            # This span is inside the context of an existing trace
-            existing_zipkin_attrs = get_zipkin_attrs()
-            if existing_zipkin_attrs:
-                self.zipkin_attrs = ZipkinAttrs(
-                    trace_id=existing_zipkin_attrs.trace_id,
-                    span_id=generate_random_64bit_string(),
-                    parent_span_id=existing_zipkin_attrs.span_id,
-                    flags=existing_zipkin_attrs.flags,
-                    is_sampled=existing_zipkin_attrs.is_sampled,
+            # If sample_rate is set, we need to (re)generate a trace context.
+            # If zipkin_attrs (trace context) were passed in as argument there are
+            # 2 possibilities:
+            # is_sampled = False --> we keep the same trace_id but re-roll the dice
+            #                        for is_sampled.
+            # is_sampled = True  --> we don't want to stop sampling halfway through
+            #                        a sampled trace, so we do nothing.
+            # If no zipkin_attrs were passed in, we generate new ones and start a
+            # new trace.
+            if self.sample_rate is not None:
+
+                # If this trace is not sampled, we re-roll the dice.
+                if self.zipkin_attrs and not self.zipkin_attrs.is_sampled:
+                    # This will be the root span of the trace, so we should
+                    # set timestamp and duration.
+                    report_root_timestamp = True
+                    self.zipkin_attrs = create_attrs_for_span(
+                        sample_rate=self.sample_rate,
+                        trace_id=self.zipkin_attrs.trace_id,
+                        use_128bit_trace_id=self.use_128bit_trace_id,
+                    )
+
+                # If zipkin_attrs was not passed in, we simply generate new
+                # zipkin_attrs to start a new trace.
+                elif not self.zipkin_attrs:
+                    # This will be the root span of the trace, so we should
+                    # set timestamp and duration.
+                    report_root_timestamp = True
+                    self.zipkin_attrs = create_attrs_for_span(
+                        sample_rate=self.sample_rate,
+                        use_128bit_trace_id=self.use_128bit_trace_id,
+                    )
+
+            if self.firehose_handler and not self.zipkin_attrs:
+                # If it has gotten here, the only thing that is
+                # causing a trace is the firehose. So we force a trace
+                # with sample rate of 0
+                report_root_timestamp = True
+                self.zipkin_attrs = create_attrs_for_span(
+                    sample_rate=0.0,
+                    use_128bit_trace_id=self.use_128bit_trace_id,
                 )
+        else:
+            # If zipkin_attrs was not passed in, we check if there's already a
+            # trace context in _context_stack.
+            if not self.zipkin_attrs:
+                existing_zipkin_attrs = self._context_stack.get()
+                # If there's an existing context, let's create new zipkin_attrs
+                # with that context as parent.
+                if existing_zipkin_attrs:
+                    self.zipkin_attrs = ZipkinAttrs(
+                        trace_id=existing_zipkin_attrs.trace_id,
+                        span_id=generate_random_64bit_string(),
+                        parent_span_id=existing_zipkin_attrs.span_id,
+                        flags=existing_zipkin_attrs.flags,
+                        is_sampled=existing_zipkin_attrs.is_sampled,
+                    )
 
         # If zipkin_attrs are not set up by now, that means this span is not
         # configured to perform logging itself, and it's not in an existing
@@ -277,58 +397,35 @@ class zipkin_span(object):
         if not self.zipkin_attrs:
             return self
 
-        push_zipkin_attrs(self.zipkin_attrs)
+        self._context_stack.push(self.zipkin_attrs)
         self.do_pop_attrs = True
 
         self.start_timestamp = time.time()
 
-        if self.perform_logging:
+        if self._is_local_root_span:
             # Don't set up any logging if we're not sampling
-            if not self.zipkin_attrs.is_sampled:
+            if not self.zipkin_attrs.is_sampled and not self.firehose_handler:
                 return self
             endpoint = create_endpoint(self.port, self.service_name, self.host)
-            client_context = set(self.include) == {'client'}
-            self.log_handler = ZipkinLoggerHandler(self.zipkin_attrs)
             self.logging_context = ZipkinLoggingContext(
                 self.zipkin_attrs,
                 endpoint,
-                self.log_handler,
                 self.span_name,
                 self.transport_handler,
                 report_root_timestamp or self.report_root_timestamp_override,
+                self._span_storage,
+                self.service_name,
                 binary_annotations=self.binary_annotations,
                 add_logging_annotation=self.add_logging_annotation,
-                client_context=client_context,
+                client_context=self.kind == Kind.CLIENT,
                 max_span_batch_size=self.max_span_batch_size,
+                firehose_handler=self.firehose_handler,
+                encoding=self.encoding,
             )
             self.logging_context.start()
-            self.logging_configured = True
-            return self
-        else:
-            # In the sampled case, patch the ZipkinLoggerHandler.
-            if self.zipkin_attrs.is_sampled:
-                # Be defensive about logging setup. Since ZipkinAttrs are local to
-                # the thread, multithreaded frameworks can get in strange states.
-                # The logging is not going to be correct in these cases, so we set
-                # a flag that turns off logging on __exit__.
-                try:
-                    # Assume there's only a single handler, since all logging
-                    # should be set up in this package.
-                    log_handler = zipkin_logger.handlers[0]
-                except IndexError:
-                    return self
-                # Make sure it's not a NullHandler or something
-                if not isinstance(log_handler, ZipkinLoggerHandler):
-                    return self
-                # Put span ID on logging handler.
-                self.log_handler = zipkin_logger.handlers[0]
-                # Store the old parent_span_id, probably None, in case we have
-                # nested zipkin_spans
-                self.old_parent_span_id = self.log_handler.parent_span_id
-                self.log_handler.parent_span_id = self.zipkin_attrs.span_id
-                self.logging_configured = True
+            self._span_storage.set_transport_configured(configured=True)
 
-            return self
+        return self
 
     def __exit__(self, _exc_type, _exc_value, _exc_traceback):
         self.stop(_exc_type, _exc_value, _exc_traceback)
@@ -339,17 +436,21 @@ class zipkin_span(object):
         popped off. The actual logging of spans depends on sampling and that
         the logging was correctly set up.
         """
-        if self.do_pop_attrs:
-            pop_zipkin_attrs()
 
-        if not self.logging_configured:
+        if self.do_pop_attrs:
+            self._context_stack.pop()
+
+        # If no transport is configured, there's no reason to create a new Span.
+        # This also helps avoiding memory leaks since without a transport nothing
+        # would pull spans out of _span_storage.
+        if not self._span_storage.is_transport_configured():
             return
 
         # Add the error annotation if an exception occurred
         if any((_exc_type, _exc_value, _exc_traceback)):
-            error_msg = '{0}: {1}'.format(_exc_type.__name__, _exc_value)
+            error_msg = u'{0}: {1}'.format(_exc_type.__name__, _exc_value)
             self.update_binary_annotations({
-                zipkin_core.ERROR: error_msg,
+                ERROR_KEY: error_msg,
             })
 
         # Logging context is only initialized for "root" spans of the local
@@ -358,38 +459,32 @@ class zipkin_span(object):
         if self.logging_context:
             self.logging_context.stop()
             self.logging_context = None
+            self._span_storage.set_transport_configured(configured=False)
             return
 
         # If we've gotten here, that means that this span is a child span of
         # this context's root span (i.e. it's a zipkin_span inside another
         # zipkin_span).
         end_timestamp = time.time()
+        # If self.duration is set, it means the user wants to override it
+        if self.duration:
+            duration = self.duration
+        else:
+            duration = end_timestamp - self.start_timestamp
 
-        self.log_handler.parent_span_id = self.old_parent_span_id
-
-        # We are simulating a full two-part span locally, so set cs=sr and ss=cr
-        full_annotations = {
-            'cs': self.start_timestamp,
-            'sr': self.start_timestamp,
-            'ss': end_timestamp,
-            'cr': end_timestamp,
-        }
-        # But we filter down if we only want to emit some of the annotations
-        filtered_annotations = {
-            k: v for k, v in full_annotations.items()
-            if k in self.annotation_filter
-        }
-
-        self.annotations.update(filtered_annotations)
-
-        self.log_handler.store_local_span(
-            span_name=self.span_name,
-            service_name=self.service_name,
-            annotations=self.annotations,
-            binary_annotations=self.binary_annotations,
-            sa_binary_annotations=self.sa_binary_annotations,
+        self._span_storage.append(SpanBuilder(
+            trace_id=self.zipkin_attrs.trace_id,
+            name=self.span_name,
+            parent_id=self.zipkin_attrs.parent_span_id,
             span_id=self.zipkin_attrs.span_id,
-        )
+            timestamp=self.timestamp if self.timestamp else self.start_timestamp,
+            duration=duration,
+            annotations=self.annotations,
+            tags=self.binary_annotations,
+            kind=self.kind,
+            service_name=self.service_name,
+            sa_endpoint=self.sa_endpoint,
+        ))
 
     def update_binary_annotations(self, extra_annotations):
         """Updates the binary annotations for the current span.
@@ -398,8 +493,6 @@ class zipkin_span(object):
         """
         if not self.zipkin_attrs:
             return
-        if not self.zipkin_attrs.is_sampled:
-            return
         if not self.logging_context:
             # This is not the root span, so binary annotations will be added
             # to the log handler when this span context exits.
@@ -429,10 +522,10 @@ class zipkin_span(object):
         :param host: Host address of the destination
         :type host: str
         """
-        if not self.zipkin_attrs or not self.zipkin_attrs.is_sampled:
+        if not self.zipkin_attrs:
             return
 
-        if 'client' not in self.include:
+        if self.kind != Kind.CLIENT:
             # TODO: trying to set a sa binary annotation for a non-client span
             # should result in a logged error
             return
@@ -442,22 +535,35 @@ class zipkin_span(object):
             service_name=service_name,
             host=host,
         )
-        sa_binary_annotation = create_binary_annotation(
-            key=zipkin_core.SERVER_ADDR,
-            value=SERVER_ADDR_VAL,
-            annotation_type=zipkin_core.AnnotationType.BOOL,
-            host=sa_endpoint,
-        )
         if not self.logging_context:
-            self.sa_binary_annotations.append(sa_binary_annotation)
+            if self.sa_endpoint is not None:
+                raise ValueError('SA annotation already set.')
+            self.sa_endpoint = sa_endpoint
         else:
-            self.logging_context.sa_binary_annotations.append(sa_binary_annotation)
+            if self.logging_context.sa_endpoint is not None:
+                raise ValueError('SA annotation already set.')
+            self.logging_context.sa_endpoint = sa_endpoint
+
+    def override_span_name(self, name):
+        """Overrides the current span name.
+
+        This is useful if you don't know the span name yet when you create the
+        zipkin_span object. i.e. pyramid_zipkin doesn't know which route the
+        request matched until the function wrapped by the context manager
+        completes.
+
+        :param name: New span name
+        :type name: str
+        """
+        self.span_name = name
+        if self.logging_context:
+            self.logging_context.span_name = name
 
 
 def _validate_args(kwargs):
-    if 'include' in kwargs:
+    if 'kind' in kwargs:
         raise ValueError(
-            '"include" is not valid in this context. '
+            '"kind" is not valid in this context. '
             'You probably want to use zipkin_span()'
         )
 
@@ -475,7 +581,7 @@ class zipkin_client_span(zipkin_span):
         """
         _validate_args(kwargs)
 
-        kwargs['include'] = ('client',)
+        kwargs['kind'] = Kind.CLIENT
         super(zipkin_client_span, self).__init__(*args, **kwargs)
 
 
@@ -492,7 +598,7 @@ class zipkin_server_span(zipkin_span):
         """
         _validate_args(kwargs)
 
-        kwargs['include'] = ('server',)
+        kwargs['kind'] = Kind.SERVER
         super(zipkin_server_span, self).__init__(*args, **kwargs)
 
 
@@ -512,6 +618,8 @@ def create_attrs_for_span(
     :param span_id: Optional 16-character hex string representing a span_id.
                     If this is None, a random span_id will be generated.
     :type span_id: str
+    :param use_128bit_trace_id: If true, generate 128-bit trace_ids
+    :type use_128bit_trace_id: boolean
     """
     # Calculate if this trace is sampled based on the sample rate
     if trace_id is None:
@@ -535,20 +643,21 @@ def create_attrs_for_span(
     )
 
 
-def create_http_headers_for_new_span():
+def create_http_headers_for_new_span(context_stack=None):
     """
     Generate the headers for a new zipkin span.
 
     .. note::
 
-        If the method is not called from within a zipkin_trace conext,
+        If the method is not called from within a zipkin_trace context,
         empty dict will be returned back.
 
     :returns: dict containing (X-B3-TraceId, X-B3-SpanId, X-B3-ParentSpanId,
                 X-B3-Flags and X-B3-Sampled) keys OR an empty dict.
     """
-    zipkin_attrs = get_zipkin_attrs()
-
+    if context_stack is None:
+        context_stack = ThreadLocalStack()
+    zipkin_attrs = context_stack.get()
     if not zipkin_attrs:
         return {}
 
